function iwmrrf()
% iwmrrf (iwmrrf companion code)
% by Christian Himpe, 2015 ( http://gramian.de )
% released under BSD 2-Clause License ( opensource.org/licenses/BSD-2-Clause )
%

POINTS = 6;
SYSDIM = 4.^(1:POINTS)

TIMINGS = zeros(POINTS,1+1+3+1+1+1+3+1+3+1+1);
TIMINGS(:,2) = NaN;
TIMINGS(:,6) = NaN;
TIMINGS(:,8) = NaN;
TIMINGS(:,12) = NaN;
TIMINGS(:,16) = NaN;

NORMS = zeros(POINTS,2+1+1+2);

J = 1;
O = J;

global CUSTOM_ODE;

for I=1:POINTS

    N = 4^I;

    if(N>400),
        T = [0.0,0.001,0.1];
    else,
        T = [0.0,0.01,1.0];
    end

    L = (T(3)-T(1))/T(2);
    UT = [ones(J,1)./T(2),zeros(J,L-1)];

    A = -gallery('lehmer',N);
    B = rand(N,J);
    C = rand(O,N);

    X = ones(N,1);
    P = rand(N,1);

    NON = @(x,u,p) A*tanh(p.*x) + B*u;
    OUT = @(x,u,p) C*x;

    tic;
    WXa = emgr_old(NON,OUT,[J,N,O],T,'j',P,0,1,0,X);
    TIMINGS(I,1) = toc;

%% TWO-STEP RUNGE-KUTTA

    for K=1:10
        tic;
        Y = rk3(NON,OUT,T(2),L,X,UT,P);
        TIMINGS(I,3) = TIMINGS(I,3) + toc;

        tic;
        Y = irk3(NON,OUT,T(2),L,X,UT,P);
        TIMINGS(I,4) = TIMINGS(I,4) + toc;
    end
    TIMINGS(I,3) = TIMINGS(I,3)/10;
    TIMINGS(I,4) = TIMINGS(I,4)/10;

    CUSTOM_ODE = @irk3;
    tic;
    WXb = emgr_old(NON,OUT,[J,N,O],T,'j',P,[0,0,0,0,0,0,0,0,0,0,0,-1],1,0,X);
    TIMINGS(I,5) = toc;

    NORMS(I,1) = norm(WXa{1}-WXb{1},'fro');
    NORMS(I,2) = norm(WXa{2}-WXb{2},'fro');

%% GENERALIZED TRANSPOSITIONS

    tic;
    WXc = emgr_dot(NON,OUT,[J,N,O],T,'j',P,0,1,0,X);
    TIMINGS(I,7) = toc;


%% APPROXIMATE INVERSE

    for K=1:10
        tic;
        WXinv = pinv(WXa{1});
        TIMINGS(I,9) = TIMINGS(I,9) + toc;

        tic;
        WXinv = ainv(WXa{1});
        TIMINGS(I,10) = TIMINGS(I,10) + toc;
    end
    TIMINGS(I,9) = TIMINGS(I,9)/10;
    TIMINGS(I,10) = TIMINGS(I,10)/10;

    tic;
    WXd = emgr_inv(NON,OUT,[J,N,O],T,'j',P,0,1,0,X);
    TIMINGS(I,11) = toc;

    NORMS(I,3) = norm(WXa{2}-WXd{2},'fro');

%% LANCZOS SVD

    tic;
    [U,D1,V] = svds(WXa{1},N);
    TIMINGS(I,13) = toc;

    tic;
    [U,D2] = eigs(WXa{1}*WXa{1}',N);
    TIMINGS(I,14) = toc;

    tic;
    [U,D3] = lsvd(WXa{1},N);
    TIMINGS(I,15) = toc;

    D1 = diag(D1);
    D1(N) = 0;
    NORMS(I,4) = norm(D1-D3,2);

%% NEW

    tic;
    WXf = emgr(NON,OUT,[J,N,O],T,'j',P,0,1,0,X);
    TIMINGS(I,17) = toc;

    NORMS(I,5) = norm(WXa{1}-WXf{1},'fro');
    NORMS(I,6) = norm(WXa{2}-WXf{2},'fro');

    fprintf('#');

end


%% TERMINAL OUTPUT

TIMINGS
NORMS

save('iwmrrf.mat','POINTS','TIMINGS','NORMS');

end


function x = ainv(m)

    d = 1.0./diag(m);
    n = numel(d);
    x = bsxfun(@times,m,-d);
    x = bsxfun(@times,x,d');
    x(1:n+1:end) = d;
end


function x = rk3(f,g,h,T,z,u,p)

    if(isnumeric(g) && g==1), g = @(x,u,p) x; end;

    k1 = h*f(z,u(:,1),p);
    k2 = h*f(z + 0.5*k1,u(:,1),p);
    k3 = h*f(z + 0.75*k2,u(:,1),p);
    z = z + (2.0/9.0)*k1 + (1.0/3.0)*k2 + (4.0/9.0)*k3; % Ralston RK3
    x(:,1) = g(z,u(:,1),p);

    x(end,T) = 0; % preallocate trajectory

    for t=2:T % 3rd Order Improved Runge-Kutta IRK3
        k1 = h*f(z,u(:,t),p);
        k2 = h*f(z + 0.5*k1,u(:,t),p);
        k3 = h*f(z + 0.75*k2,u(:,t),p);
        z = z + (2.0/9.0)*k1 + (1.0/3.0)*k2 + (4.0/9.0)*k3; % Ralston RK3
        x(:,t) = g(z,u(:,t),p);
    end;
end


function x = irk3(f,g,h,T,z,u,p)

    if(isnumeric(g) && g==1), g = @(x,u,p) x; end;

    k1 = h*f(z,u(:,1),p); % 2nd Order Midpoint RK2 for starting value
    k2 = h*f(z + 0.5*k1,u(:,1),p);
    z = z + k2;
    x(:,1) = g(z,u(:,1),p);

    x(end,T) = 0; % preallocate trajectory

    k1 = h*f(z,u(:,2),p); % 2nd start value (improves impulse response)
    k2 = h*f(z + 0.5*k1,u(:,2),p);
    z = z + k2;
    x(:,2) = g(z,u(:,2),p);

    for t=3:T % 3rd Order Improved Runge-Kutta IRK3
        l1 = h*f(z,u(:,t),p);
        l2 = h*f(z + 0.5*l1,u(:,t),p);
        z = z + (2.0/3.0)*l1 + (1.0/3.0)*k1 + (5.0/6.0)*(l2 - k2);
        x(:,t) = g(z,u(:,t),p);
        k1 = l1;
        k2 = l2;
    end;
end
